# ruff: noqa: UP006
from __future__ import annotations

import warnings
from functools import cached_property, partial
from typing import TYPE_CHECKING

import graphene
from graphene.relay.connection import connection_adapter, page_info_adapter
from graphene.types.argument import to_arguments
from graphene.utils.str_converters import to_camel_case, to_snake_case
from graphene_django.settings import graphene_settings
from graphene_django.utils.utils import DJANGO_FILTER_INSTALLED, maybe_queryset
from graphql_relay.connection.array_connection import offset_to_cursor

from .ast import get_underlying_type
from .compiler import OptimizationCompiler, optimize
from .prefetch_hack import fetch_in_context
from .settings import optimizer_settings
from .utils import calculate_queryset_slice, is_optimized
from .validators import validate_pagination_args

if TYPE_CHECKING:
    from django.db import models
    from django.db.models import Model
    from django.db.models.manager import Manager
    from graphene.relay.connection import Connection
    from graphql_relay import EdgeType
    from graphql_relay.connection.connection import ConnectionType

    from .types import DjangoObjectType
    from .typing import (
        Any,
        Callable,
        ConnectionResolver,
        ExpressionKind,
        GQLInfo,
        Iterable,
        ModelResolver,
        ObjectTypeInput,
        Optional,
        QuerySetResolver,
        Type,
        Union,
        UnmountedTypeInput,
    )

__all__ = [
    "DjangoConnectionField",
    "DjangoListField",
    "RelatedField",
    "AnnotatedField",
    "MultiField",
]


class RelatedField(graphene.Field):
    """Field for `to-one` related models with default resolvers."""

    def __init__(
        self,
        type_: ObjectTypeInput,
        /,
        *,
        field_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a related field for the given type.

        :param type_: DjangoObjectType the related field is for.
                      This can also be a dot import path to the object type,
                      or a callable that returns the object type.
        :param field_name: The name of the model field or related accessor this related field is for.
                           Only needed if the field name on the ObjectType this field is
                           defined on is different from the field name on the model.
        :param kwargs: Extra arguments passed to `graphene.types.field.Field`.
        """
        if kwargs.pop("reverse", None) is not None:  # pragma: no cover
            msg = (
                "The `reverse` argument is no longer required, and should be removed. "
                "This will become an error in the future."
            )
            warnings.warn(msg, category=DeprecationWarning, stacklevel=1)

        self.field_name = field_name
        super().__init__(type_, **kwargs)

    def wrap_resolve(self, parent_resolver: ModelResolver) -> ModelResolver:
        # Allow user defined resolvers to override the default behavior.
        if not isinstance(parent_resolver, partial):
            return parent_resolver
        return self.related_resolver

    def related_resolver(self, root: models.Model, info: GQLInfo) -> Optional[models.Model]:
        field_name = self.field_name or to_snake_case(info.field_name)
        # Related object should be optimized to the root model.
        related_instance: Optional[models.Model] = getattr(root, field_name, None)
        if related_instance is None:  # pragma: no cover
            return None
        self.underlying_type.run_instance_checks(related_instance, info)
        return related_instance

    @cached_property
    def underlying_type(self) -> type[DjangoObjectType]:
        return get_underlying_type(self.type)


class FilteringMixin:
    # Subclasses should implement the following:
    underlying_type: type[DjangoObjectType]

    no_filters: bool = False
    """Should filterset filters be disabled for this field?"""

    @property
    def args(self) -> dict[str, graphene.Argument]:
        return to_arguments(getattr(self, "_base_args", None) or {}, self.filtering_args)

    @args.setter
    def args(self, args: dict[str, Any]) -> None:
        # noinspection PyAttributeOutsideInit
        self._base_args = args

    @cached_property
    def filtering_args(self) -> Optional[dict[str, graphene.Argument]]:
        if not DJANGO_FILTER_INSTALLED or self.no_filters:  # pragma: no cover
            return None

        from graphene_django.filter.utils import get_filtering_args_from_filterset

        filterset_class = getattr(self.underlying_type._meta, "filterset_class", None)
        if filterset_class is None:
            return None
        return get_filtering_args_from_filterset(filterset_class, self.underlying_type)


class DjangoListField(FilteringMixin, graphene.Field):
    """DjangoListField that also supports filtering."""

    def __init__(
        self,
        type_: ObjectTypeInput,
        /,
        *,
        no_filters: bool = False,
        field_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a list field for the given type.

        :param type_: DjangoObjectType the list field is for.
                      This can also be a dot import path to the object type,
                      or a callable that returns the object type.
        :param no_filters: Should filterset filters be disabled for this field?
        :param field_name: The name of the model field or related accessor this list field is for.
                           Only needed if the field name on the ObjectType this field is
                           defined on is different from the field name on the model.
        :param kwargs: Extra arguments passed to `graphene.types.field.Field`.
        """
        self.no_filters = no_filters
        self.field_name = field_name
        if isinstance(type_, graphene.NonNull):  # pragma: no cover
            type_ = type_.of_type
        super().__init__(graphene.List(graphene.NonNull(type_)), **kwargs)

    def wrap_resolve(self, parent_resolver: QuerySetResolver) -> QuerySetResolver:
        self.resolver = parent_resolver
        return self.list_resolver

    def list_resolver(self, root: Any, info: GQLInfo, **kwargs: Any) -> models.QuerySet:
        result = self.resolver(root, info, **kwargs)
        queryset = self.to_queryset(result)
        queryset = self.underlying_type.get_queryset(queryset, info)

        max_complexity: Optional[int] = getattr(self.underlying_type._meta, "max_complexity", None)
        return optimize(queryset, info, max_complexity=max_complexity)

    def to_queryset(self, iterable: Union[models.QuerySet, Manager, None]) -> models.QuerySet:
        # Default resolver can return a Manager-instance or None.
        if iterable is None:
            iterable = self.model._default_manager
        return maybe_queryset(iterable)

    @cached_property
    def underlying_type(self) -> type[DjangoObjectType]:
        return get_underlying_type(self.type)

    @cached_property
    def model(self) -> type[models.Model]:
        return self.underlying_type._meta.model


class DjangoConnectionField(FilteringMixin, graphene.Field):
    """DjangoConnectionField for Django models that works for both filtered and non-filtered Relay-nodes."""

    def __init__(
        self,
        type_: ObjectTypeInput,
        /,
        *,
        max_limit: Optional[int] = ...,
        no_filters: bool = False,
        field_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a connection field for the given type.

        :param type_: DjangoObjectType the connection is for.
                      This can also be a dot import path to the object type,
                      or a callable that returns the object type.
        :param max_limit: Maximum number of items that can be requested in a single query for this connection.
                          Set to None to disable the limit.
        :param no_filters: Should filterset filters be disabled for this field?
        :param field_name: The name of the model field or related accessor this connection is for.
                           Only needed if the field name on the ObjectType this field is
                           defined on is different from the field name on the model.
        :param kwargs: Extra arguments passed to `graphene.types.field.Field`.
        """
        # Maximum number of items that can be requested in a single query for this connection.
        # Set to None to disable the limit.
        self.max_limit = max_limit if max_limit is not ... else graphene_settings.RELAY_CONNECTION_MAX_LIMIT
        self.no_filters = no_filters
        self.field_name = field_name

        # Default inputs for a connection field
        kwargs.setdefault("first", graphene.Int())
        kwargs.setdefault("last", graphene.Int())
        kwargs.setdefault("offset", graphene.Int())
        kwargs.setdefault("after", graphene.String())
        kwargs.setdefault("before", graphene.String())

        super().__init__(type_, **kwargs)

    def wrap_resolve(self, parent_resolver: QuerySetResolver) -> ConnectionResolver:
        self.resolver = parent_resolver
        return self.connection_resolver

    def connection_resolver(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionType:
        pagination_args = validate_pagination_args(
            first=kwargs.pop("first", None),
            last=kwargs.pop("last", None),
            offset=kwargs.pop("offset", None),
            after=kwargs.pop("after", None),
            before=kwargs.pop("before", None),
            max_limit=self.max_limit,
        )

        # Call the ObjectType's "resolve_{field_name}" method if it exists.
        # Otherwise, call the default resolver (usually `dict_or_attr_resolver`).
        result = self.resolver(root, info, **kwargs)
        queryset = self.to_queryset(result)
        queryset = self.underlying_type.get_queryset(queryset, info)

        max_complexity: Optional[int] = getattr(self.underlying_type._meta, "max_complexity", None)

        # Note if the queryset has already been optimized.
        already_optimized = is_optimized(queryset)

        optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
        if optimizer is not None:
            queryset = optimizer.optimize_queryset(queryset)

        # Queryset optimization contains filtering, so we count after optimization.
        pagination_args["size"] = count = (
            queryset.count()
            if not already_optimized
            # Prefetch(..., to_attr=...) will return a list of models.
            # TODO: This might be wrong.
            else len(queryset)
            if isinstance(queryset, list)
            # If this is a nested connection field, prefetch queryset models should have been
            # annotated with the queryset count (pick it from the first one).
            else getattr(
                next(iter(getattr(queryset, "_result_cache", []) or []), None),
                optimizer_settings.PREFETCH_COUNT_KEY,
                0,  # QuerySet result cache is empty -> count is 0.
            )
        )
        cut = calculate_queryset_slice(**pagination_args)

        # Prefetch queryset has already been sliced.
        if not already_optimized:
            queryset = queryset[cut]

        edges: list[EdgeType] = [
            # Create a connection from the sliced queryset.
            self.connection_type.Edge(node=value, cursor=offset_to_cursor(cut.start + index))
            for index, value in enumerate(fetch_in_context(queryset))
        ]

        connection = connection_adapter(
            cls=self.connection_type,
            edges=edges,
            pageInfo=page_info_adapter(
                startCursor=edges[0].cursor if edges else None,
                endCursor=edges[-1].cursor if edges else None,
                hasPreviousPage=cut.start > 0,
                hasNextPage=cut.stop < count,
            ),
        )
        connection.iterable = queryset
        connection.length = count
        return connection

    def to_queryset(self, iterable: Union[models.QuerySet, Manager, None]) -> models.QuerySet:
        # Default resolver can return a Manager-instance or None.
        if iterable is None:
            iterable = self.model._default_manager
        return maybe_queryset(iterable)

    @cached_property
    def type(self) -> Union[Type[Connection], graphene.NonNull]:  # pragma: no cover
        type_ = super().type
        non_null = isinstance(type_, graphene.NonNull)
        if non_null:
            type_ = type_.of_type

        connection_type: Optional[Type[Connection]] = type_._meta.connection
        if connection_type is None:
            msg = f"The type {type_.__name__} doesn't have a connection"
            raise ValueError(msg)

        if non_null:
            return graphene.NonNull(connection_type)
        return connection_type

    @cached_property
    def connection_type(self) -> Type[Connection]:  # pragma: no cover
        type_ = self.type
        if isinstance(type_, graphene.NonNull):
            return type_.of_type
        return type_

    @cached_property
    def underlying_type(self) -> Type[DjangoObjectType]:
        return self.connection_type._meta.node

    @cached_property
    def model(self) -> Type[models.Model]:
        return self.underlying_type._meta.model


class AnnotatedField(graphene.Field):
    """Field for resolving Django ORM expressions that the optimizer will annotate to the queryset."""

    def __init__(
        self,
        type_: UnmountedTypeInput,
        /,
        expression: ExpressionKind,
        aliases: Optional[dict[str, ExpressionKind]] = None,
        **kwargs: Any,
    ) -> None:
        self.expression = expression
        self.aliases = aliases
        super().__init__(type_, **kwargs)

    def __set_name__(self, owner: type[DjangoObjectType], name: str) -> None:
        self.name = to_camel_case(name)

    def wrap_resolve(self, parent_resolver: Callable[..., Any]) -> Callable[..., Any]:
        # `parent_resolver` is either a `resolve_{self.name}` method defined
        # on the owner class, or a partial of `dict_or_attr_resolver`.
        self.resolver = parent_resolver
        return self.annotation_resolver

    def annotation_resolver(self, root: Model, info: GQLInfo, **kwargs: Any) -> Any:
        return self.resolver(root, info, **kwargs)


class MultiField(graphene.Field):
    """Field that requires multiple model fields to resolve. Does not support related lookups."""

    def __init__(self, type_: UnmountedTypeInput, /, fields: Iterable[str], **kwargs: Any) -> None:
        self.fields = fields
        super().__init__(type_, **kwargs)

    def __set_name__(self, owner: type[DjangoObjectType], name: str) -> None:
        self.name = to_camel_case(name)

    def wrap_resolve(self, parent_resolver: Callable[..., Any]) -> Callable[..., Any]:
        # `parent_resolver` is either a `resolve_{self.name}` method defined
        # on the owner class, or a partial of `dict_or_attr_resolver`.
        self.resolver = parent_resolver
        return self.multi_field_resolver

    def multi_field_resolver(self, root: Model, info: GQLInfo, **kwargs: Any) -> Any:
        return self.resolver(root, info, **kwargs)



from __future__ import annotations

import dataclasses
from copy import copy
from typing import TYPE_CHECKING

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Model, Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.functions import RowNumber
from graphene_django.registry import get_global_registry
from graphene_django.settings import graphene_settings

from .ast import get_model_field
from .filter_info import get_filter_info
from .prefetch_hack import _register_for_prefetch_hack
from .settings import optimizer_settings
from .utils import (
    SubqueryCount,
    add_slice_to_queryset,
    calculate_queryset_slice,
    calculate_slice_for_queryset,
    mark_optimized,
    optimizer_logger,
)
from .validators import validate_pagination_args

if TYPE_CHECKING:
    from .types import DjangoObjectType
    from .typing import Any, ExpressionKind, GQLInfo, GraphQLFilterInfo, Optional, TModel, ToManyField


__all__ = [
    "QueryOptimizer",
]


@dataclasses.dataclass
class CompilationResults:
    only_fields: list[str] = dataclasses.field(default_factory=list)
    select_related: list[str] = dataclasses.field(default_factory=list)
    prefetch_related: list[Prefetch | str] = dataclasses.field(default_factory=list)


class QueryOptimizer:
    """Creates optimized queryset based on the optimization data found by the OptimizationCompiler."""

    def __init__(self, model: type[Model] | None, info: GQLInfo, name: Optional[str] = None) -> None:
        self.model = model
        self.info = info
        self.only_fields: list[str] = []
        self.related_fields: list[str] = []
        self.aliases: dict[str, ExpressionKind] = {}
        self.annotations: dict[str, ExpressionKind] = {}
        self.select_related: dict[str, QueryOptimizer] = {}
        self.prefetch_related: dict[str, QueryOptimizer] = {}
        self.total_count: bool = False
        self.name = name

    def optimize_queryset(
        self,
        queryset: QuerySet[TModel],
        *,
        filter_info: Optional[GraphQLFilterInfo] = None,
    ) -> QuerySet[TModel]:
        """
        Add the optimizations in this optimizer to the given queryset.

        :param queryset: QuerySet to optimize.
        :param filter_info: Additional filtering info to use for the optimization.
        """
        if filter_info is None:
            filter_info = get_filter_info(self.info, queryset.model)

        results = self.compile(filter_info=filter_info)

        if filter_info is not None and filter_info.get("filterset_class") is not None:
            filterset = filter_info["filterset_class"](
                data=self.process_filters(filter_info["filters"]),
                queryset=queryset,
                request=self.info.context,
            )
            if not filterset.is_valid():  # pragma: no cover
                raise ValidationError(filterset.form.errors.as_json())

            queryset = filterset.qs

        queryset = self.get_filtered_queryset(queryset)

        if results.select_related:
            queryset = queryset.select_related(*results.select_related)
        if results.prefetch_related:
            queryset = queryset.prefetch_related(*results.prefetch_related)
        if not optimizer_settings.DISABLE_ONLY_FIELDS_OPTIMIZATION and (results.only_fields or self.related_fields):
            queryset = queryset.only(*results.only_fields, *self.related_fields)
        if self.aliases:
            queryset = queryset.alias(**self.aliases)
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        mark_optimized(queryset)
        return queryset

    def compile(self, *, filter_info: GraphQLFilterInfo) -> CompilationResults:
        results = CompilationResults(only_fields=self.only_fields.copy())

        for name, optimizer in self.select_related.items():
            # Promote select related to prefetch related if any annotations are needed.
            if optimizer.annotations:
                self.compile_prefetch(name, optimizer, results, filter_info)
            else:
                self.compile_select(name, optimizer, results, filter_info)

        for attr, optimizer in self.prefetch_related.items():
            self.compile_prefetch(attr, optimizer, results, filter_info)

        return results

    def compile_select(
        self,
        name: str,
        optimizer: QueryOptimizer,
        results: CompilationResults,
        filter_info: GraphQLFilterInfo,
    ) -> None:
        results.select_related.append(name)
        nested_results = optimizer.compile(filter_info=filter_info)
        results.only_fields.extend(f"{name}{LOOKUP_SEP}{only}" for only in nested_results.only_fields)
        results.select_related.extend(f"{name}{LOOKUP_SEP}{select}" for select in nested_results.select_related)
        for prefetch in nested_results.prefetch_related:
            if isinstance(prefetch, Prefetch):
                prefetch.add_prefix(name)
                results.prefetch_related.append(prefetch)

    def compile_prefetch(
        self,
        attr: str,
        optimizer: QueryOptimizer,
        results: CompilationResults,
        filter_info: GraphQLFilterInfo,
    ) -> None:
        if optimizer.model is None:  # generic foreign keys
            results.prefetch_related.append(optimizer.name)
            return

        filter_info = filter_info.get("children", {}).get(optimizer.name, {})
        queryset = optimizer.model._default_manager.all()
        queryset = optimizer.optimize_queryset(queryset, filter_info=filter_info)
        queryset = optimizer.paginate_prefetch_queryset(self.model, queryset, optimizer.name, filter_info=filter_info)
        to_attr = attr if attr != optimizer.name else None
        results.prefetch_related.append(Prefetch(optimizer.name, queryset, to_attr=to_attr))

    def paginate_prefetch_queryset(
        self,
        parent_model: type[Model],
        queryset: QuerySet[TModel],
        name: str,
        filter_info: GraphQLFilterInfo,
    ) -> QuerySet[TModel]:
        """Paginate prefetch queryset based on the given filter info after it has been filtered."""
        # Only paginate nested connection fields.
        if not filter_info.get("is_connection", False):
            return queryset

        field: Optional[ToManyField] = get_model_field(parent_model, name)
        if field is None:  # pragma: no cover
            msg = f"Cannot find field {name!r} on model {parent_model.__name__!r}. Cannot optimize nested pagination."
            optimizer_logger.warning(msg)
            return queryset

        remote_field = field.remote_field
        field_name = remote_field.name if isinstance(field, models.ManyToManyField) else remote_field.attname

        order_by: list[str] = (
            # Use the `order_by` from the filter info, if available
            [x for x in filter_info.get("filters", {}).get("order_by", "").split(",") if x]
            # Use the model's `Meta.ordering` if no `order_by` is given
            or copy(queryset.model._meta.ordering)
            # No ordering if neither is available
            or []
        )

        pagination_args = validate_pagination_args(
            after=filter_info.get("filters", {}).get("after"),
            before=filter_info.get("filters", {}).get("before"),
            offset=filter_info.get("filters", {}).get("offset"),
            first=filter_info.get("filters", {}).get("first"),
            last=filter_info.get("filters", {}).get("last"),
            max_limit=filter_info.get("max_limit", graphene_settings.RELAY_CONNECTION_MAX_LIMIT),
        )

        if self.total_count or pagination_args.get("last") is not None or pagination_args.get("size") is None:
            # If the query asks for total count for a nested connection field,
            # or is trying to limit the number of items from the end of the list,
            # or the user has set the `max_size` for the field to None (=no limit),
            # annotate the models in the queryset with the total count for each partition.
            # This is optional, since there is a performance impact due to needing
            # to use a subquery for each partition.
            queryset = queryset.annotate(
                **{
                    optimizer_settings.PREFETCH_COUNT_KEY: SubqueryCount(
                        queryset.filter(**{field_name: models.OuterRef(field_name)}),
                    ),
                },
            )

        # Don't limit the queryset if no pagination arguments are given (and field `max_size=None`)
        if all(value is None for value in pagination_args.values()):  # pragma: no cover
            return queryset

        if pagination_args.get("last") is not None or pagination_args.get("size") is None:
            queryset = calculate_slice_for_queryset(queryset, **pagination_args)
        else:
            cut = calculate_queryset_slice(**pagination_args)
            queryset = add_slice_to_queryset(queryset, start=models.Value(cut.start), stop=models.Value(cut.stop))

        _register_for_prefetch_hack(self.info, field)

        return (
            queryset
            # Add a row number to the queryset, and limit the rows for each
            # partition based on the given pagination arguments.
            .alias(
                **{
                    optimizer_settings.PREFETCH_PARTITION_INDEX: (
                        models.Window(
                            expression=RowNumber(),
                            partition_by=models.F(field_name),
                            order_by=order_by,
                        )
                        - models.Value(1)  # Start from zero.
                    )
                },
            ).filter(
                **{
                    f"{optimizer_settings.PREFETCH_PARTITION_INDEX}__gte": models.F(
                        optimizer_settings.PREFETCH_SLICE_START
                    ),
                    f"{optimizer_settings.PREFETCH_PARTITION_INDEX}__lt": models.F(
                        optimizer_settings.PREFETCH_SLICE_STOP
                    ),
                }
            )
        )

    def get_filtered_queryset(self, queryset: QuerySet[TModel]) -> QuerySet[TModel]:
        object_type: Optional[DjangoObjectType] = get_global_registry().get_type_for_model(queryset.model)
        if callable(getattr(object_type, "filter_queryset", None)):
            return object_type.filter_queryset(queryset, self.info)  # type: ignore[union-attr]
        return queryset  # pragma: no cover

    def process_filters(self, input_data: dict[str, Any]) -> dict[str, Any]:
        from graphene_django.filter.fields import convert_enum

        return {key: convert_enum(value) for key, value in input_data.items()}



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

import django_filters
from django.utils.module_loading import import_string
from graphene_django.utils import maybe_queryset

from .settings import optimizer_settings

if TYPE_CHECKING:
    from django.db import models


__all__ = [
    "FilterSet",
    "create_filterset",
    "default_filterset_class",
]


class FilterSet(django_filters.FilterSet):
    @property
    def qs(self) -> models.QuerySet:
        if not hasattr(self, "_qs"):
            # Override the default `qs` property to include this.
            # It's needed to ensure that the `queryset._result_cache`
            # is not cleared when using `queryset.all()`
            qs = maybe_queryset(self.queryset)
            if self.is_bound:
                # ensure form validation before filtering
                # noinspection PyStatementEffect
                self.errors  # noqa: B018
                qs = self.filter_queryset(qs)
            # noinspection PyAttributeOutsideInit
            self._qs = qs
        return self._qs


@cache
def default_filterset_class() -> type[FilterSet]:
    if optimizer_settings.DEFAULT_FILTERSET_CLASS:  # pragma: no cover
        return import_string(optimizer_settings.DEFAULT_FILTERSET_CLASS)
    return FilterSet


def create_filterset(
    model: type[models.Model],
    fields: dict[str, list[str]],
) -> type[FilterSet]:
    name = f"{model._meta.object_name}FilterSet"
    meta = type("Meta", (), {"model": model, "fields": fields})
    filterset_class = default_filterset_class()
    return type(name, (filterset_class,), {"Meta": meta})  # type: ignore[return-type]



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db.models import ForeignKey, Manager, ManyToOneRel, Model, QuerySet
from graphene.utils.str_converters import to_snake_case
from graphene_django.utils import maybe_queryset

from .ast import GraphQLASTWalker
from .errors import OptimizerError
from .optimizer import QueryOptimizer
from .prefetch_hack import fetch_in_context
from .settings import optimizer_settings
from .utils import is_optimized, optimizer_logger

if TYPE_CHECKING:
    import graphene
    from django.db import models
    from graphene.types.definitions import GrapheneObjectType
    from graphql import FieldNode

    from .typing import PK, GQLInfo, Optional, TModel, ToManyField, ToOneField, Union


__all__ = [
    "optimize",
    "optimize_single",
    "OptimizationCompiler",
]


def optimize(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    max_complexity: Optional[int] = None,
) -> QuerySet[TModel]:
    """Optimize the given queryset according to the field selections received in the GraphQLResolveInfo."""
    optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
    if optimizer is not None:
        queryset = optimizer.optimize_queryset(queryset)
        fetch_in_context(queryset)

    return queryset


def optimize_single(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    pk: PK,
    max_complexity: Optional[int] = None,
) -> Optional[TModel]:
    """Optimize the given queryset for a single model instance by its primary key."""
    optimizer = OptimizationCompiler(info, max_complexity=max_complexity).compile(queryset)
    if optimizer is None:  # pragma: no cover
        return queryset.filter(pk=pk).first()

    queryset = optimizer.optimize_queryset(queryset.filter(pk=pk))
    fetch_in_context(queryset)

    # Shouldn't use .first(), as it can apply additional ordering, which would cancel the optimization.
    # The queryset should have the right model instance, since we started by filtering by its pk,
    # so we can just pick that out of the result cache (if it hasn't been filtered out).
    return next(iter(queryset), None)


class OptimizationCompiler(GraphQLASTWalker):
    """Class for compiling SQL optimizations based on the given query."""

    def __init__(self, info: GQLInfo, max_complexity: Optional[int] = None) -> None:
        """
        Initialize the optimization compiler with the query info.

        :param info: The GraphQLResolveInfo containing the query AST.
        :param max_complexity: How many 'select_related' and 'prefetch_related' table joins are allowed.
                               Used to protect from malicious queries.
        """
        self.max_complexity = max_complexity or optimizer_settings.MAX_COMPLEXITY
        self.optimizer: QueryOptimizer = None  # type: ignore[assignment]
        self.to_attr: Optional[str] = None
        super().__init__(info)

    def compile(self, queryset: Union[QuerySet, Manager, list[Model]]) -> Optional[QueryOptimizer]:
        """
        Compile optimizations for the given queryset.

        :return: QueryOptimizer instance that can perform any needed optimization,
                 or None if queryset is already optimized.
        :raises OptimizerError: Something went wrong during the optimization process.
        """
        queryset = maybe_queryset(queryset)
        # If prior optimization has been done already, return early.
        if is_optimized(queryset):
            return None

        # Setup initial state.
        self.model = queryset.model
        self.optimizer = QueryOptimizer(model=queryset.model, info=self.info)

        # Walk the query AST to compile the optimizations.
        try:
            self.run()

        # Allow known errors to be raised.
        except OptimizerError:  # pragma: no cover
            raise

        # Raise unknown errors if not allowed to skip optimization on error.
        except Exception as error:  # noqa: BLE001  # pragma: no cover
            optimizer_logger.warning("Something went wrong during the optimization process.", exc_info=error)
            if not optimizer_settings.SKIP_OPTIMIZATION_ON_ERROR:
                raise
            return None

        return self.optimizer

    def increase_complexity(self) -> None:
        super().increase_complexity()
        if self.complexity > self.max_complexity:
            msg = f"Query complexity exceeds the maximum allowed of {self.max_complexity}"
            raise OptimizerError(msg)

    def handle_normal_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field: models.Field) -> None:
        self.optimizer.only_fields.append(field.get_attname())

    def handle_to_one_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
        related_model: type[Model],
    ) -> None:
        name = related_field.get_cache_name() or related_field.name
        optimizer = QueryOptimizer(model=related_model, info=self.info, name=name)

        if isinstance(related_field, GenericForeignKey):
            self.optimizer.prefetch_related[name] = optimizer
        else:
            self.optimizer.select_related[name] = optimizer

        if isinstance(related_field, ForeignKey):
            self.optimizer.related_fields.append(related_field.attname)

        if isinstance(related_field, GenericForeignKey):
            self.optimizer.related_fields.append(related_field.ct_field)
            self.optimizer.related_fields.append(related_field.fk_field)

        with self.use_optimizer(optimizer):
            super().handle_to_one_field(field_type, field_node, related_field, related_model)

    def handle_to_many_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
        related_model: type[Model],
    ) -> None:
        name = related_field.get_cache_name() or related_field.name
        key = self.to_attr if self.to_attr is not None else name
        self.to_attr = None

        optimizer = QueryOptimizer(model=related_model, info=self.info, name=name)
        self.optimizer.prefetch_related[key] = optimizer

        if isinstance(related_field, ManyToOneRel):
            optimizer.related_fields.append(related_field.field.attname)

        if isinstance(related_field, GenericRelation):
            optimizer.related_fields.append(related_field.object_id_field_name)
            optimizer.related_fields.append(related_field.content_type_field_name)

        with self.use_optimizer(optimizer):
            super().handle_to_many_field(field_type, field_node, related_field, related_model)

    def handle_total_count(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        self.optimizer.total_count = True

    def handle_custom_field(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)
        field: Optional[graphene.Field] = field_type.graphene_type._meta.fields.get(field_name)
        if field is None:  # pragma: no cover
            msg = (
                f"Field '{field_node.name.value}' not found from object type '{field_type.graphene_type}'. "
                f"Cannot optimize custom field."
            )
            optimizer_logger.warning(msg)
            return None

        # `RelatedField`, `DjangoListField` and `DjangoConnectionField` can define a
        # 'field_name' attribute to specify the actual model field name.
        actual_field_name: Optional[str] = getattr(field, "field_name", None)
        if actual_field_name is not None:
            self.to_attr = field_name
            return self.handle_model_field(field_type, field_node, actual_field_name)

        from .fields import AnnotatedField, MultiField

        if isinstance(field, AnnotatedField):
            self.optimizer.annotations[to_snake_case(field.name)] = field.expression
            if field.aliases is not None:
                self.optimizer.aliases.update(field.aliases)
            return None

        if isinstance(field, MultiField):
            self.optimizer.only_fields.extend(field.fields)
            return None

        return None  # pragma: no cover

    @contextlib.contextmanager
    def use_optimizer(self, optimizer: QueryOptimizer) -> None:
        orig_optimizer = self.optimizer
        try:
            self.optimizer = optimizer
            yield
        finally:
            self.optimizer = orig_optimizer



from __future__ import annotations

import contextlib
from contextlib import suppress

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Field, ForeignKey, Model
from graphene import Connection, ObjectType, PageInfo
from graphene.relay.node import AbstractNode
from graphene.types.definitions import GrapheneObjectType, GrapheneUnionType
from graphene.utils.str_converters import to_snake_case
from graphene_django import DjangoObjectType
from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLField,
    GraphQLOutputType,
    InlineFragmentNode,
    SelectionNode,
)
from graphql.execution.execute import get_field_def

from .errors import OptimizerError
from .settings import optimizer_settings
from .typing import (
    GRAPHQL_BUILTIN,
    GQLInfo,
    ModelField,
    Optional,
    ToManyField,
    ToOneField,
    TypeGuard,
    Union,
    overload,
)

__all__ = [
    "GraphQLASTWalker",
]


GrapheneType = Union[GrapheneObjectType, GrapheneUnionType]
Selections = tuple[SelectionNode, ...]


class GraphQLASTWalker:
    """Class for walking the GraphQL AST and handling the different nodes."""

    def __init__(self, info: GQLInfo, model: Optional[type[Model]] = None) -> None:
        self.info = info
        self.complexity: int = 0
        self.model: type[Model] = model

    def increase_complexity(self) -> None:
        self.complexity += 1

    def run(self) -> None:
        selections: Selections = self.info.field_nodes  # type: ignore[assignment]
        field_type: GrapheneObjectType = self.info.parent_type  # type: ignore[assignment]
        return self.handle_selections(field_type, selections)

    def handle_selections(self, field_type: GrapheneType, selections: Selections) -> None:
        for selection in selections:
            if isinstance(selection, FieldNode):
                self.handle_field_node(field_type, selection)

            elif isinstance(selection, FragmentSpreadNode):
                self.handle_fragment_spread(field_type, selection)

            elif isinstance(selection, InlineFragmentNode):
                self.handle_inline_fragment(field_type, selection)

            else:  # pragma: no cover
                msg = f"Unhandled selection node: '{selection}'"
                raise OptimizerError(msg)

    def handle_field_node(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type: type[ObjectType] = field_type.graphene_type

        if self.info.parent_type == field_type:
            return self.handle_query_class(field_type, field_node)

        if issubclass(graphene_type, Connection):
            return self.handle_connection(field_type, field_node)

        if is_edge(field_type):
            return self.handle_edge(field_type, field_node)

        if issubclass(graphene_type, PageInfo):  # pragma: no cover
            return self.handle_page_info(field_type, field_node)

        if issubclass(graphene_type, ObjectType):
            return self.handle_object_type(field_type, field_node)

        msg = f"Unhandled graphene type: '{graphene_type}'"  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover

    def handle_query_class(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_object_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)
        if is_graphql_builtin(field_name):
            return self.handle_graphql_builtin(field_type, field_node)

        graphene_type: type[ObjectType] = field_type.graphene_type

        if issubclass(graphene_type, DjangoObjectType):
            return self.handle_model_field(field_type, field_node, field_name)
        return self.handle_plain_object_type(field_type, field_node)

    def handle_graphql_builtin(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_plain_object_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_model_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field_name: str) -> None:
        model: type[Model] = field_type.graphene_type._meta.model
        field = get_model_field(model, field_name)

        if field is None:
            with self.use_model(model):
                return self.handle_custom_field(field_type, field_node)

        if not field.is_relation or is_foreign_key_id(field, field_node):
            with self.use_model(model):
                return self.handle_normal_field(field_type, field_node, field)

        if is_to_one(field):
            related_model = get_related_model(field, model)
            with self.use_model(field.model):
                return self.handle_to_one_field(field_type, field_node, field, related_model)

        if is_to_many(field):
            related_model = get_related_model(field, model)
            with self.use_model(model):
                return self.handle_to_many_field(field_type, field_node, field, related_model)

        msg = f"Unhandled field: '{field.name}'"  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover

    def handle_custom_field(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_normal_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field: Field) -> None: ...

    def handle_to_one_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
        related_model: type[Model],
    ) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        self.increase_complexity()
        return self.handle_selections(graphene_type, selections)

    def handle_to_many_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
        related_model: type[Model],
    ) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        self.increase_complexity()
        return self.handle_selections(graphene_type, selections)

    def handle_connection(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        if field_node.name.value == optimizer_settings.TOTAL_COUNT_FIELD:
            return self.handle_total_count(field_type, field_node)

        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_edge(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_total_count(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_page_info(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_fragment_spread(self, field_type: GrapheneObjectType, fragment_spread: FragmentSpreadNode) -> None:
        name = fragment_spread.name.value
        fragment_definition = self.info.fragments[name]
        selections = get_selections(fragment_definition)
        return self.handle_selections(field_type, selections)

    def handle_inline_fragment(self, field_type: GrapheneUnionType, inline_fragment: InlineFragmentNode) -> None:
        fragment_type = get_fragment_type(field_type, inline_fragment)
        fragment_model: type[Model] = fragment_type.graphene_type._meta.model
        if fragment_model != self.model:
            return None

        selections = get_selections(inline_fragment)
        return self.handle_selections(fragment_type, selections)

    def get_graphene_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> GrapheneType:
        graphql_field = get_field_def(self.info.schema, field_type, field_node)
        return get_underlying_type(graphql_field.type)

    @contextlib.contextmanager
    def use_model(self, model: type[Model]) -> GraphQLASTWalker:
        orig_model = self.model
        try:
            self.model = model
            yield
        finally:
            self.model = orig_model


@overload
def get_underlying_type(
    field_type: type[GraphQLOutputType],
) -> type[Union[DjangoObjectType, GrapheneObjectType]]: ...  # pragma: no cover


@overload
def get_underlying_type(
    field_type: GraphQLOutputType,
) -> Union[DjangoObjectType, GrapheneObjectType]: ...  # pragma: no cover


def get_underlying_type(field_type):
    while hasattr(field_type, "of_type"):
        field_type = field_type.of_type
    return field_type


def get_selections(field_node: Union[FieldNode, FragmentDefinitionNode, InlineFragmentNode]) -> Selections:
    return getattr(field_node.selection_set, "selections", ())


def is_edge(field_type: GrapheneObjectType) -> bool:
    # Edge-classes are created by `graphene.relay.connection.get_edge_class`,
    # which means that we cannot check against the EdgeBase class directly.
    return all(field in field_type.fields for field in ["node", "cursor"]) and field_type.name.endswith("Edge")


def is_connection(graphql_field: Union[GraphQLField, GrapheneObjectType]) -> bool:
    return issubclass(getattr(graphql_field, "graphene_type", type(None)), Connection)


def is_node(graphql_field: GraphQLField) -> bool:
    return issubclass(getattr(getattr(graphql_field.resolve, "func", None), "__self__", type(None)), AbstractNode)


def is_graphql_builtin(field_name: str) -> bool:
    return field_name.lower() in GRAPHQL_BUILTIN


def is_foreign_key_id(field: Field, field_node: FieldNode) -> bool:
    return isinstance(field, ForeignKey) and field.get_attname() == to_snake_case(field_node.name.value)


def is_to_many(field: Field) -> TypeGuard[ToManyField]:
    return bool(field.one_to_many or field.many_to_many)


def is_to_one(field: Field) -> TypeGuard[ToOneField]:
    return bool(field.many_to_one or field.one_to_one)


def get_fragment_type(field_type: GrapheneUnionType, inline_fragment: InlineFragmentNode) -> GrapheneObjectType:
    fragment_type_name = inline_fragment.type_condition.name.value
    gen = (t for t in field_type.types if t.name == fragment_type_name)
    fragment_type: Optional[GrapheneObjectType] = next(gen, None)

    if fragment_type is None:  # pragma: no cover
        msg = f"Fragment type '{fragment_type_name}' not found in union '{field_type}'"
        raise OptimizerError(msg)

    return fragment_type


def get_related_model(related_field: Union[ToOneField, ToManyField], model: type[Model]) -> type[Model]:
    related_model = related_field.related_model
    if related_model == "self":  # pragma: no cover
        return model
    return related_model  # type: ignore[return-value]


def get_model_field(model: type[Model], field_name: str) -> Optional[ModelField]:
    if field_name == "pk":
        model_field: ModelField = model._meta.pk
        return model_field

    with suppress(FieldDoesNotExist):
        model_field: ModelField = model._meta.get_field(field_name)
        return model_field

    # Field might be a reverse many-related field without `related_name`, in which case
    # the `model._meta.fields_map` will store the relation without the "_set" suffix.
    if field_name.endswith("_set"):
        with suppress(FieldDoesNotExist):
            model_field: ModelField = model._meta.get_field(field_name.removesuffix("_set"))
            if is_to_many(model_field):
                return model_field

    return None



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from graphene.utils.str_converters import to_snake_case

from query_optimizer.ast import GraphQLASTWalker, get_selections

if TYPE_CHECKING:
    from django.db import models
    from graphene.types.definitions import GrapheneObjectType
    from graphql import FieldNode

    from .typing import Any, GQLInfo, Optional, ToManyField, ToOneField


__all__ = [
    "get_field_selections",
]


def get_field_selections(info: GQLInfo, model: Optional[type[models.Model]] = None) -> list[Any]:
    """Compile filter information included in the GraphQL query."""
    compiler = FieldSelectionCompiler(info, model)
    compiler.run()
    return compiler.field_selections[0][to_snake_case(info.field_name)]


class FieldSelectionCompiler(GraphQLASTWalker):
    """Class for compiling filtering information from a GraphQL query."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.field_selections: list[Any] = []
        super().__init__(*args, **kwargs)

    def handle_query_class(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        with self.child_selections(field_node):
            return super().handle_query_class(field_type, field_node)

    def handle_graphql_builtin(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        self.field_selections.append(to_snake_case(field_node.name.value))  # pragma: no cover

    def handle_plain_object_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        selections = get_selections(field_node)

        if not selections:
            self.field_selections.append(to_snake_case(field_node.name.value))
            return None

        graphene_type = self.get_graphene_type(field_type, field_node)
        with self.child_selections(field_node):
            return self.handle_selections(graphene_type, selections)

    def handle_normal_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field: models.Field) -> None:
        self.field_selections.append(to_snake_case(field_node.name.value))

    def handle_custom_field(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        self.field_selections.append(to_snake_case(field_node.name.value))  # pragma: no cover

    def handle_to_one_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
        related_model: type[models.Model],
    ) -> None:
        with self.child_selections(field_node):
            return super().handle_to_many_field(field_type, field_node, related_field, related_model)

    def handle_to_many_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
        related_model: type[models.Model],
    ) -> None:
        with self.child_selections(field_node):
            return super().handle_to_one_field(field_type, field_node, related_field, related_model)

    @contextlib.contextmanager
    def child_selections(self, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)
        selections: list[Any] = []
        orig_selections = self.field_selections
        try:
            self.field_selections = selections
            yield
        finally:
            self.field_selections = orig_selections
            if selections:
                self.field_selections.append({field_name: selections})



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    ContextManager,
    Generator,
    Generic,
    Hashable,
    Iterable,
    Literal,
    NamedTuple,
    Optional,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.db import models
from graphene.types.unmountedtype import UnmountedType
from graphene_django import DjangoObjectType
from graphene_django.types import DjangoObjectTypeOptions
from graphql_relay import ConnectionType

# New in version 3.10
try:
    from typing import ParamSpec, TypeAlias, TypeGuard
except ImportError:
    from typing_extensions import ParamSpec, TypeAlias, TypeGuard


from django.core.handlers.wsgi import WSGIRequest
from django.db.models import (
    Field,
    ForeignKey,
    ForeignObject,
    ForeignObjectRel,
    Manager,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    QuerySet,
)
from graphql import GraphQLResolveInfo

if TYPE_CHECKING:
    from django.contrib.auth.models import AnonymousUser, User
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
    from django.db.models.sql import Query
    from django_filters import FilterSet

__all__ = [
    "Any",
    "Callable",
    "Collection",
    "ConnectionResolver",
    "ContextManager",
    "Expr",
    "ExpressionKind",
    "GQLInfo",
    "GRAPHQL_BUILTIN",
    "Generator",
    "Generic",
    "GraphQLFilterInfo",
    "Hashable",
    "Iterable",
    "Literal",
    "ModelField",
    "ModelResolver",
    "NamedTuple",
    "ObjectTypeInput",
    "OptimizedDjangoOptions",
    "Optional",
    "PK",
    "ParamSpec",
    "QuerySetResolver",
    "ToManyField",
    "ToOneField",
    "Type",
    "TypeGuard",
    "TypeVar",
    "TypedDict",
    "Union",
    "UnmountedTypeInput",
    "cast",
    "overload",
]


TModel = TypeVar("TModel", bound=Model)
PK: TypeAlias = Any
ModelField: TypeAlias = Union[Field, ForeignObjectRel, "GenericForeignKey"]
ToManyField: TypeAlias = Union["GenericRelation", ManyToManyField, ManyToOneRel, ManyToManyRel]
ToOneField: TypeAlias = Union["GenericRelation", ForeignObject, ForeignKey, OneToOneField]
AnyUser: TypeAlias = Union["User", "AnonymousUser"]
QuerySetResolver: TypeAlias = Callable[..., Union[QuerySet, Manager, None]]
ModelResolver: TypeAlias = Callable[..., Union[Model, None]]
ConnectionResolver: TypeAlias = Callable[..., ConnectionType]
ObjectTypeInput: TypeAlias = Union[type[DjangoObjectType], str, Callable[[], type[DjangoObjectType]]]
UnmountedTypeInput: TypeAlias = Union[type[UnmountedType], str, Callable[[], type[UnmountedType]]]
Expr: TypeAlias = Union[models.Expression, models.F, models.Q]


GRAPHQL_BUILTIN = (
    "__typename",
    "__schema",
    "__type",
    "__typekind",
    "__field",
    "__inputvalue",
    "__enumvalue",
    "__directive",
)


class UserHintedWSGIRequest(WSGIRequest):
    user: AnyUser


class GQLInfo(GraphQLResolveInfo):
    context: UserHintedWSGIRequest


class OptimizedDjangoOptions(DjangoObjectTypeOptions):
    max_complexity: int


class GraphQLFilterInfo(TypedDict, total=False):
    name: str
    filters: dict[str, Any]
    children: dict[str, GraphQLFilterInfo]
    filterset_class: Optional[type[FilterSet]]
    is_connection: bool
    is_node: bool
    max_limit: Optional[int]


class ExpressionKind(Protocol):
    def resolve_expression(
        self,
        query: Query,
        allow_joins: bool,  # noqa: FBT001
        reuse: set[str] | None,
        summarize: bool,  # noqa: FBT001
        for_save: bool,  # noqa: FBT001
    ) -> ExpressionKind: ...
